require 'dot_notation'

class JiraIssue
  include SLATracker
  include TriageTracker
  # The FIVE_P_DATE variable exists for tickets created before
  # 2015-03-06 15:55:16 -0500. Not used for any tickets after that.
  # Tickets created after above date use differnt SLA measurements.
  FIVE_P_DATE                   = Time.at(142_567_531_6).freeze
  STORY_POINTS_FIELD            = 'customfield_10007'.freeze
  TRIAGER_FIELD                 = 'customfield_14100'.freeze
  SLA_TRIAGED_AT_FIELD          = 'customfield_14400'.freeze
  OVER_TRIAGE_SLA_FIELD         = 'customfield_14200'.freeze
  OVER_TRIAGE_SLA_FIELD_YES     = '12500'.freeze
  OVER_TRIAGE_SLA_FIELD_NO      = '12501'.freeze
  API_OVER_TRIAGE_SLA_FIELD_YES = {"self"=>"https://optoro.atlassian.net/rest/api/2/customFieldOption/12500", "value"=>"Yes", "id"=>"12500"}.freeze
  API_OVER_TRIAGE_SLA_FIELD_NO  = {"self"=>"https://optoro.atlassian.net/rest/api/2/customFieldOption/12501", "value"=>"No", "id"=>"12501"}.freeze
  SLA_DUE_TIME_FIELD            = 'customfield_12900'.freeze
  SLA_DUE_WARNING_FIELD         = 'customfield_13401'.freeze
  SLA_DUE_WARNING_FIELD_SET     = '12100'.freeze
  SLA_CLOSED_AT_FIELD           = 'customfield_14000'.freeze
  OVER_SLA_FIELD                = 'customfield_12901'.freeze
  OVER_SLA_FIELD_YES            = '11901'.freeze
  OVER_SLA_FIELD_NO             = '11900'.freeze
  TOTAL_TIME_OVER_SLA_FIELD     = 'customfield_14201'.freeze
  API_OVER_SLA_FIELD_YES        = {"self"=>"https://optoro.atlassian.net/rest/api/2/customFieldOption/11901", "value"=>"Yes", "id"=>"11901"}.freeze
  API_OVER_SLA_FIELD_NO         = {"self"=>"https://optoro.atlassian.net/rest/api/2/customFieldOption/11900", "value"=>"No", "id"=>"11900"}.freeze
  AFFECTED_USER_GROUP_FIELD     = 'customfield_13602'.freeze

  attr_accessor :key, :created_at, :type, :priority_name, :status, :orig,
                :url, :project, :story_points, :raw_sprints, :description,
                :summary, :labels, :assigned_team, :changelog, :sla_due_time,
                :sla_due_warning, :over_sla,:affected_user_group,
                :cant_reproduce, :in_progress, :not_accepted, :requester_denied,
                :requester_review, :resolved, :merged, :merged_st,
                :on_production, :deploy, :completed, :reverify, :pr_review,
                :triaged, :sla_triaged_at, :over_triage_sla, :triager,
                :verified, :sla_closed_at, :total_time_over_sla

  def initialize(issue)
    issue.extend(DotNotation)
    self.key                 = issue.dot 'key'
    self.created_at          = Time.parse(issue.dot('fields.created'))
    self.type                = issue.dot 'fields.issuetype.name'
    self.priority_name       = issue.dot 'fields.priority.name'
    self.status              = issue.dot('fields.status.name').downcase
    self.project             = issue.dot 'fields.project.key'
    self.description         = issue.dot 'fields.description'
    self.summary             = issue.dot 'fields.summary'
    self.labels              = issue.dot 'fields.labels'
    self.assigned_team       = issue.dot 'fields.customfield_12601.value'
    self.story_points        = issue.dot "fields.#{STORY_POINTS_FIELD}"
    self.orig                = issue
    self.url                 = "https://optoro.atlassian.net/browse/#{key}"
    self.raw_sprints         = issue.dot 'fields.customfield_10007'
    self.sla_triaged_at      = Time.parse(issue.dot("fields.#{SLA_TRIAGED_AT_FIELD}")) if issue.dot("fields.#{SLA_TRIAGED_AT_FIELD}").present?
    self.over_triage_sla     = issue.dot "fields.#{OVER_TRIAGE_SLA_FIELD}"
    self.triager             = issue.dot "fields.#{TRIAGER_FIELD}.name"
    self.sla_due_time        = Time.parse(issue.dot("fields.#{SLA_DUE_TIME_FIELD}")) if issue.dot("fields.#{SLA_DUE_TIME_FIELD}").present?
    self.sla_due_warning     = issue.dot "fields.#{SLA_DUE_WARNING_FIELD}"
    self.sla_closed_at       = Time.parse(issue.dot("fields.#{SLA_CLOSED_AT_FIELD}")) if issue.dot("fields.#{SLA_CLOSED_AT_FIELD}").present?
    self.over_sla            = issue.dot "fields.#{OVER_SLA_FIELD}"
    self.total_time_over_sla = issue.dot "fields.#{TOTAL_TIME_OVER_SLA_FIELD}"
    self.affected_user_group = issue.dot "fields.#{AFFECTED_USER_GROUP_FIELD}"

    inv = self.class.custom_time_fields.invert
    inv.each do |name, key|
      x = name.gsub(/'/,'').gsub(/ +/, '_').downcase.gsub(/support_xt_/,'')
      v = issue.dot "fields.#{key}"
      v = Time.parse(v) if v
      x = "pr_review" if x == "review" # rename
      send("#{x}=", v)
    end
  end

  def jiralicious_issue
    Jiralicious::Issue.find(key)
  end

  def self.find(query, offset = 0)
    ask = 100
    result = Jiralicious.search(query, start_at: offset, max_results: ask)
    issues = result.issues_raw
    issues.map! { |i| new(i) }
    if issues.count == ask && result.num_results > offset + ask
      issues += find(query, offset + ask)
    end
    issues
  rescue Jiralicious::JqlError => e
    $stderr.puts "Error: #{e}"
    []
  end

  def self.custom_time_fields
    Hash[[
      [10812, "Support xt Can't Reproduce"],
      [10808, 'Support xt In Progress'],
      [10809, 'Support xt Not Accepted'],
      [10817, 'Support xt Requester Denied'],
      [10814, 'Support xt Requester Review'],
      [13101, 'Support xt Merged'],
      [13102, 'Support xt On Production'],
      [10816, 'Support xt Resolved'],
      [10815, 'Support xt Reverify'],
      [10813, 'Support xt Review'],
      [10810, 'Support xt Triaged'],
      [10811, 'Support xt Verified']
    ].map{|k,v| ["customfield_#{k}", v]}]
  end

  def open?
    !closed?
    # ["unverified", "triaged", "verified",
    #  "in progress", "pr_review", "pr review"].include? status
  end

  def closed?
    [ "requester review",
      "merged",
      "merged_st",
      "on production",
      "resolved",
      "not accepted",
      "can't reproduce",
      "deploy",
      "completed"
    ].include? status
  end

  def closed_at
    [ requester_review,
      merged,
      merged_st,
      on_production,
      resolved,
      not_accepted,
      cant_reproduce,
      deploy,
      completed
    ].compact.min
  end

  def bug?
    type == "Bug / Outage"
  end

  def operational?
    type == "Operational Request"
  end

  def feature?
    type == "Feature Enhancement"
  end

  def research?
    type == "Research"
  end

  def affected_user_groups
    affected_user_group.map {|user_group| user_group["value"]} unless affected_user_group.nil?
  end

  def affects_warehouse_users?
    return false if affected_user_groups.nil?
    affected_user_groups.include?("WH Users")
  end

  def affects_customers?
    return false if affected_user_groups.nil?
    affected_user_groups.include?("Customers")
  end

  def affects_client_corporate_users?
    return false if affected_user_groups.nil?
    affected_user_groups.include?("Client Corporate Users")
  end

  def affects_internal_optoro_users?
    return false if affected_user_groups.nil?
    affected_user_groups.include?("Internal Optoro Users")
  end

  def priority
    case priority_name
    when /P1/
      :p1
    when /P2/
      :p2
    when /P3/
      :p3
    when /P4/
      :p4
    else
      :p5
    end
  end

  def time_to_qa
    if triaged && (verified || cant_reproduce || in_progress)
      if verified
        verified - triaged
      elsif cant_reproduce
        cant_reproduce - triaged
      else
        in_progress - triaged
      end
    elsif triaged
      Time.now - triaged
    else
      nil
    end
  end

  def sprints
    if raw_sprints
      raw_sprints.map {|s| Hash[s.match(/\[(.*)\]/)[1].split(',').map{|f|f.split('=')}]}
    else
      []
    end
  end

  def changelog
    return @changelog if @changelog
    changelog = Jiralicious::Issue.fetch(key: key, body_to_params: true, body: {expand: 'changelog'})
    @changelog = changelog["changelog"]["histories"]
  end

  def changelog_simple
    changelog.map { |c|
      c["items"].map { |i|
        i.slice("field", "fromString","toString")
         .merge({
           "created" => c["created"],
           "author_out" => c["author"]["name"] || ""
         })
      }
    }.flatten
  end

  def status_changes
    status_changes = changelog_simple.select{|c| c["field"] == "status"}
    start = created_at
    author = orig["fields"]["reporter"].andand["name"] || ""
    status_changes.each { |c|
      c["age_hours"] = (Time.parse(c["created"]).to_i - start.to_i)/60.0/60
      c["started"] = start
      start = Time.parse(c["created"])
      cur_author = author
      author = c["author_out"]
      c["author_in"] = cur_author
    }
    statuses = status_changes.map { |s|
      {
        status: s["fromString"],
        start: s["started"],
        end: Time.parse(s["created"]),
        author_in: s["author_in"],
        author_out: s["author_out"],
        age_hours: s["age_hours"]
      }
    }
    statuses << {
      status: status,
      start: start,
      author_in: author
    }

  end

  def status_ages
    ages = {}
    status_changes.each{|c| ages[c[:status]] = ages.fetch(c[:status], 0) + c[:age_hours].to_r.round(2).to_f}
    ages
  end

  def status_duration_during(status, start_date, end_date)
    ages = {}
    status_changes.each{|c| ages[c["fromString"]] = ages.fetch(c["fromString"],0) + c["age_hours"]}
    status_changes.each{|c| ages[c[:status]] = ages.fetch(c[:status], 0) + c[:age_hours].to_r.round(2).to_f}
    last_status = status_changes.last
    if last_status
      ages[last_status["toString"]] =
        ages.fetch(last_status["toString"], 0) +
        ((Time.now.to_f - Time.parse(last_status["created"]).to_f)/60.0/60)
    end
    ages
  end
end
